/*
 * Copyright (c) 2001 Takayuki, Hagihara.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted.
 *
 * Do not suppress this copyright notice.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: exefilter.c,v 1.16 2002/08/21 09:34:34 hagi Exp $
 *
 * ver 2.1 08/21/2001
 *  $content-type added for .exe with audio/x-wav
 *  correct $envelope-x typo
 *
 * ver 2.0b3 12/18/2001
 *  correct open backupfile twice
 *
 * ver 2.0b2 10/04/2001
 *  support message/rfc822
 *  support no extention (ext is <>)
 *  correct streamfilter missing body
 *
 * ver 2.0b1 08/16/2001
 *  use rule file, RENAME, DROP, REJECT, SAVE, MSG support
 *
 * ver 1.2 08/12/2001
 *  correct body_len at line not end
 *
 * ver 1.1 08/02/2001
 *  correct multi allocate memory for bad_name.
 *
 * ver 1.0 07/29/2001
 */

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <time.h>
#ifdef MSDOS
#include <io.h>
#include "mfapi.h"
#include "exx.h"
#else
#include <sysexits.h>
#include <unistd.h>
#include <syslog.h>
#endif

#ifdef MSDOS

#define DEBUG 1

#define DEFAULT_RULE_FILE "ex.rul"

#else

#include <libmilter/mfapi.h>
#define DEFAULT_RULE_FILE "/etc/mail/exefilter.rul"

#endif

static char file_version[] = "$Id: exefilter.c,v 1.16 2002/08/21 09:34:34 hagi Exp $";
static char release_version[] = "exefilter-2.1";

typedef int bool;

#ifndef FALSE
# define FALSE	0
#endif /* ! FALSE*/
#ifndef TRUE
# define TRUE	1
#endif /* ! TRUE*/

#define MAX_BOUNDARY 6		/* max nest level of mime multipart message */

struct HEADER {
	struct HEADER *next;		/* pointer to next header */
	char	*f;
	char	*v;
	struct HEADER *pre;		/* pointer to before header (for clear) */
};

struct CHECK_RULE {
	int	rule_n;			/* rule No. */
	struct CHECK_RULE *next;	/* pointer to next rule */
	char	*ext;			/* check extension type [ext\0ext\0...ext\0\0] */
	char	*cond;			/* optional rule [rule\0...rule\0\0] */
	char	*action;		/* action if matched [action\0...action\0\0] */
};

struct mlfiPriv
{
	int	now_multipart_header;	/* now checking multipart message 1:header 0:body */
	int	multi_body_type;	/* multipart message body 0:unknown 1:text 2:html */
	int	next_body_is_header;	/* message/rf822, body is mail */

	int	body_changed;		/* body part changed by filter */
	sfsistat	final_action;		/* */

	char	*body;			/* recieved data body */
	int	body_len;		/* recived body part length */
	int	body_cur_pos;		/* checking line start position */
	int	block_start_pos;	/* multipart message current part start position */
	int	multi_header_pos;	/* multipart header start posision */

	int	multipart_level;	/* multipart messages nest level */
	char	*boundary[MAX_BOUNDARY];	/* multipart boundary string for each level */

	int	bad_name_type;		/* 0:OK 1:multipart 2:uuencode */
	char	*bad_name;		/* suppressed file name for disp */
	char	*cur_name;		/* suppressed file name of current name= */
	char	*change_extp;		/* extension's '.' posision for RENAME */

	char	*text_plain;		/* Content-type if text/plain */
	char	*text_html;		/* Content-type if text/html */

	char	*add_msg;		/* add message for this block */
	int 	msg_added;		/* already MSG added in this block */
	int	alternative;		/* message is multipart/alternative */
	int	text_add_pos;		/* text position for MSG add */
	int	text_encode;		/* text Transfer-Encording is Quoted printable? */
	int	html_add_pos;		/* html position previous </body> */
	int	html_encode;		/* html Transfer-Encording is Quoted printable? */
	int	content_type_pos;	/* content-type position of current block */
	char	*reason_ext;		/* pointer of DROP/REJECT extension */
	int	reason_line;		/* action caused line pos of rule file */

	int	save;			/* save flag 1:start 2:saving */
	FILE 	*savefp;		/* for save */
	char 	*savename;		/* backup save file name */

	struct	HEADER	 *header;	/* headers */
	struct	HEADER	 *last_header;	/* last header position */
};

#define MLFIPRIV	(stream_mode?(local_priv):((struct mlfiPriv *)smfi_getpriv(ctx)))

extern sfsistat mlfi_cleanup(SMFICTX *, sfsistat act);
extern void stream_replacebody(char *body, int body_len);
extern void stream_addheader(char *f, char *v);

#ifdef DEBUG
#define DEBUG_fprintf(x) fprintf x
FILE *debug_fp;
#else
#define DEBUG_fprintf(x)
#endif

/* cut messages for multipart body */
char *mime_cut[] = {
	"Content-Type: text/plain;\r\n\r\n== WARNING Attached file deleted by mail filter ==\r\n== deleted file is <",
	"> ==\r\n\r\n--",
	"\r\n",
};

/* cut messages for uuencode body */
char *uu_cut[] = {
	"== WARNING Attached file deleted by mail filter ==\r\n== deleted file is <",
	"> ==\r\n",
};

/* add message pre header */
char msg_hd[] = "== exefilter added message ==\r\n";


struct CHECK_RULE rules;	/* rules for check */

int stream_mode;		/* stream filter or sendmail milter */
struct mlfiPriv *local_priv;	/* priv for stream filter mode */

char *save_dir = NULL;		/* directory for SAVE original mail */
int save_expire = 168;		/* hours for remain SAVE mail not impremented */
char *id_str = NULL;		/* optional filtering header message */
int verbose = 0;		/* display readed rules */
int test_mode = 0;		/* test for REJECT/DROP .forward stream filter */
int read_lines;			/* for information, read rules and stream filter */
char *fname;			/* for information, read rules and stream filter */

/* reply code for REJECT mail */
void set_reject_reply(SMFICTX *ctx)
{
	if ( !stream_mode ) {
		smfi_setreply(ctx, "554", "5.6.1", "Don't attach executable file");
	}
}

/* output log to syslog(at milter) or to stderr(at stream filter) */
void filter_log(struct mlfiPriv *priv)
{
	char *envfrom;
	char *envrcpt;
	char *act;
	char *p;
	struct HEADER *hd;

	hd = priv->header;
	while ( hd != NULL ) {
		if ( strcmp(hd->f,"$envelope-from") == 0 ) {
			envfrom = hd->v;
		} else if ( strcmp(hd->f,"$envelope-to") == 0 ) {
			envrcpt = hd->v;
		}
		hd = hd->next;
	}
	if ( priv->final_action == SMFIS_REJECT ) {
		act = "REJECT";
	} else {
		act = "DROP";
	}
	p = priv->reason_ext;
	while ( *p && !isspace(*p) && *p != '"' && *p != '\r' && *p != '\n' && *p != ';' ) ++p;
	*p = '\0';

	if ( priv->save != 0 )  {
		if ( !stream_mode ) {
			syslog(LOG_INFO, "%s(%d): %s from=<%s> to=<%s> save=<%s>\n", act, priv->reason_line, priv->reason_ext, envfrom, envrcpt, priv->savename);
		} else {
			fprintf(stderr,"%s(%d): %s from=<%s> to=<%s> save=<%s>\n", act, priv->reason_line, priv->reason_ext, envfrom, envrcpt, priv->savename);
		}
	} else {
		if ( !stream_mode ) {
			syslog(LOG_INFO, "%s(%d): %s from=<%s> to=<%s>\n", act, priv->reason_line, priv->reason_ext, envfrom, envrcpt);
		} else {
			fprintf(stderr,"%s(%d): %s from=<%s> to=<%s>\n", act, priv->reason_line, priv->reason_ext, envfrom, envrcpt);
		}
	}
}

/* create SAVE file
**
** SAVEDIR/username@MMDD-hhmm[-nn]
** MMDDhhmm is date/time of recieved time
** nn is optional serial No. if file already exists
**/
int create_save_file(struct mlfiPriv *priv)
{
	struct HEADER *hd;
	char *p, *q;
	time_t t;
	struct tm *tt;
	int cnt;
	int fh;

	hd = priv->header;
	p = NULL;
	while ( hd != NULL ) {
		if ( strcmp(hd->f, "$envelope-to") == 0 ) {
			p = hd->v;
			break;
		}
		hd = hd->next;
	}
	if ( p == NULL ) {
		p = "Unknown";
	}
	priv->savename = malloc(strlen(save_dir)+strlen(p)+1+4+1+4+3+1);	/* <dir><name>@MMDD-hhmm-xx */
	if ( priv->savename == NULL ) return -1;
	strcpy(priv->savename, save_dir);
	q = priv->savename;
	while ( *q ) ++q;
	if ( *p == '<' ) ++p;
	while ( *p && *p != '@' && *p != '>' ) {
		*q++ = *p++;
	}
	*q++ = '@';
	time(&t);
	tt = localtime(&t);
	sprintf(q, "%02d%02d-%02d%02d", tt->tm_mon+1, tt->tm_mday, tt->tm_hour, tt->tm_min);
	cnt = 0;
	fh = open(priv->savename, O_EXCL|O_CREAT|O_WRONLY,0600);
	while ( fh == -1 ) {
		if ( ++cnt >= 100 ) return -1;
		sprintf(q, "%02d%02d-%02d%02d-%02d", tt->tm_mon+1, tt->tm_mday, tt->tm_hour, tt->tm_min, cnt);
		fh = open(priv->savename, O_EXCL|O_CREAT|O_WRONLY,0600);
	}
	priv->savefp = fdopen(fh, "w");
	return 0;
}

/* back mail if rule has SAVE
** priv->save == 1 is first call, create file and save headers and existing body
** priv->save == 2, save newer body only
*/
int backup_mail(struct mlfiPriv *priv, char *body, int len)
{
	struct HEADER *hd;

	if ( priv->save == 1 ) {
		if ( create_save_file(priv) != 0 ) return -1;
		if ( priv->savefp == NULL ) return -1;
		priv->save = 2;
		hd = priv->header;
		while ( hd != NULL ) {
			if ( hd->f[0] != '$' ) {
				fprintf(priv->savefp, "%s: %s\r\n", hd->f, hd->v);
			}
			hd = hd->next;
		}
		fprintf(priv->savefp, "\r\n");
	}
	fwrite(body, 1, len, priv->savefp);
	return 0;
}

/* output 1 rule for user infomation */
void output_rule(FILE *fp, struct CHECK_RULE *rp)
{
	char *p;

	p = rp->ext;
	while ( *p ) {			/* print extensions */
		fprintf(fp,"%s", p);
		p += strlen(p)+1;
		if ( *p ) {
			fprintf(fp,", ");
		}
	}
	p = rp->cond;
	if ( *p != '\0' ) {		/* print conditions */
		fprintf(fp," ? ");
		while ( *p ) {
			fprintf(fp,"%s:", p);
			p += strlen(p)+1;
			fprintf(fp,"\"%s\"", p);
			p += strlen(p)+1;
			if ( *p ) {
				fprintf(fp," & ");
			}
		}
	}
	p = rp->action;		/* print actions */
	fprintf(fp," : ");
	while ( *p ) {
		fprintf(fp,"%s", p);
		if ( strcasecmp(p, "MSG") == 0 ) {
			p += strlen(p)+1;
			fprintf(fp," \"%s\"", p);
		}
		p += strlen(p)+1;
		if ( *p ) {
			fprintf(fp,", ");
		}
	}
	fprintf(fp,"\n");
}

/* search buf containing str with ignoe case */
int str_in_buf(char *buf, char *str)
{
	int l;
	int pos;

	if ( buf == NULL ) return -1;
	l = strlen(str);
	pos = 0;
	while ( *buf ) {
		if ( strncasecmp(buf,str,l) == 0 ) return pos;
		++buf;
		++pos;
	}
	return -1;
}

/* check double extension "pext" in tbl */
int preext_match(char *pext, char *tbl)
{
	int l;
	char *p;

	if ( pext == NULL ) return 0;
	p = pext;
	l = 0;
	while ( *p != '.' ) {
		++l;
		++p;
	}
	while ( *tbl ) {
		if ( strncasecmp(pext, tbl, l) == 0 ) {
			if ( isspace(tbl[l]) || tbl[l] == ',' ) return 1;	/* match */
		}
		while ( *tbl && *tbl != ',' ) ++tbl;
		if ( *tbl == ',' ) ++tbl;
		while ( isspace(*tbl) ) ++tbl;
	}
	return 0;
}

/* check filename with extension to delete */
int check_bad_ext(struct mlfiPriv *priv, char *name, int hflg)
{
	char *ext;		/* extension start pos */
	char *pext;		/* double extension start pos */
	char *rext;		/* reason ext */
	char *p, *q;
	int ext_l;
	int st;
	int rev;
	struct CHECK_RULE *rp;
	struct HEADER *hp;

	DEBUG_fprintf((debug_fp, "ext_check %s", name));

	/* already DROP of DISCARD */
	if ( priv->final_action != SMFIS_CONTINUE ) return 0;

	p = name;
	ext = p;
	pext = NULL;
	while ( *p ) {		/* search extension position */
		if ( *p == '.' ) {
			pext = ext;
			ext = p;
		}
		if ( *p == '\r' || *p == '\n' || *p == ';' || *p == '"' ) break;
		++p;
	}
	if ( *ext != '.' ) {
		ext = p;
	} else {
		++ext;
	}
	if ( pext != NULL ) {
		if ( *pext != '.' ) {
			pext = NULL;
		} else {
			++pext;
		}
	}

	for ( rp = rules.next; rp != NULL; rp = rp->next) {
		p = rp->ext;
		if ( hflg == 0 ) {
			if ( *p != '-' ) continue;	/* checking body, so SKIP header only rule */
			p = p+strlen(p)+1;
		}
		if ( hflg == 1 && *p == '-' ) continue;	/* checking headers, so SKIP body rule */

		/* checking extensions */
		for ( ; *p != '\0'; p += strlen(p)+1 ) {
			if ( *p == '*' ) {
				ext_l = ext-name;
				q = ext;
				while ( *q && !isspace(*q) && *q != '"' && *q != ';' ) {
					++q;
					++ext_l;
				}
				break;
			} else if ( *p =='<' && p[1] =='>' ) {
				/* check no extention */
				if ( *ext =='\0' || isspace(*ext) || *ext =='"' || *ext == ';' ) {
					ext_l = ext-name;
					break;
				}
			}

			ext_l = strlen(p);
			if ( strncasecmp(ext, p, ext_l) == 0 ) {
				if ( ext[ext_l] && !isspace(ext[ext_l]) && ext[ext_l] != '"' && ext[ext_l] != ';' ) {
					/* ext is long than checking one */
					continue;
				}
				ext_l = ext-name+ext_l;		/* calc bad_name length */
				break;
			}
		}
		if ( *p == '\0' ) continue;		/* ext not match */
		rext = p;

#ifdef DEBUG
		DEBUG_fprintf((debug_fp, "ext match\t"));
		output_rule(debug_fp, rp);
#endif
		/* checking conditions */
		for ( p = rp->cond; *p != '\0'; p = q+strlen(q)+1 ) {
			rev = 0;
			if ( *p == '!' ) {
				rev = 1;
				++p;
			}
			hp = priv->header;
			while ( hp != NULL ) {
				if ( strcasecmp(hp->f, p) == 0 ) break;
				hp = hp->next;
			}
			q = p+strlen(p)+1;

			st = 0;
			if ( hp == NULL ) {
				if ( strcasecmp(p,"$texttype") == 0 ) {
if ( priv->text_plain != NULL ) {
	DEBUG_fprintf((debug_fp,"$texttype %s",priv->text_plain));
}
if ( priv->text_html != NULL ) {
	DEBUG_fprintf((debug_fp,"$texttype %s",priv->text_html));
}
					if ( str_in_buf(priv->text_plain, q) != -1
						|| str_in_buf(priv->text_html, q) != -1 ) {	/* match */
						st = 1;		/* $texttype OK */
					}
				} else if ( strcasecmp(p,"$preext") == 0 ) {
					if ( preext_match(pext, q) ) {
						st = 1;		/* $preext OK */
					}
				} else if ( strcasecmp(p,"$content-type") == 0 ) {
					if ( priv->content_type_pos != 0
					&& str_in_buf(&priv->body[priv->content_type_pos] , q) != -1 ) {
						st = 1;		/* $content-type OK */
					}
				}
				if ( strcmp(q,"<>") == 0 ) {
					st = 1;		/* OK this conditon */
				}
			} else if ( str_in_buf(hp->v, q) != -1 ) {	/* match */
				st = 1;		/* OK string match */
			} else if ( strcmp(q,"<>") == 0 && hp->v[0] == '\0' ) {
				st = 1;		/* OK empty header */
			}
			if ( rev == 0 && st == 0 ) {
				break;	/* BAD condition */
			} else if ( rev == 1 && st == 1 ) {
				break;	/* BAD condition */
			}
			/* OK this condition, do next */
		}
		if ( *p ) {	/* condition remain, means BAD */
			DEBUG_fprintf((debug_fp,"condition BAD\n"));

			continue;		/* conditon bad */
		}

		DEBUG_fprintf((debug_fp,"condition OK\n"));

		/* apply actions */
		p = rp->action;
		st = 0;
		while ( *p ) {
			if ( strcmp(p, "DROP") == 0 ) {
				priv->final_action = SMFIS_DISCARD;
				priv->reason_ext = rext;
				priv->reason_line = rp->rule_n;
			} else if ( strcmp(p, "REJECT") == 0 ) {
				priv->final_action = SMFIS_REJECT;
				priv->reason_ext = rext;
				priv->reason_line = rp->rule_n;
			} else if ( strcmp(p, "OK") == 0 ) {
				/* do nothing */
			} else if ( strcmp(p, "RENAME") == 0 ) {
				priv->change_extp = &ext[-1];
				priv->body_changed = 1;
			} else if ( strcmp(p, "DELETE") == 0 ) {
				st = 1;
			} else if ( strcmp(p, "SAVE") == 0 ) {
				if ( priv->save == 0 ) {
					priv->save = 1;
				}
			} else if ( strcmp(p, "MSG") == 0 ) {
				p = p+strlen(p)+1;
				if ( priv->msg_added == 0 ) {
					priv->add_msg = p;
					priv->body_changed = 1;
					priv->msg_added = 1;
				}
			}
			p = p+strlen(p)+1;
		}

		if ( st ) {
			/* at DELETE, copy filename for information */
			priv->cur_name = malloc(ext_l+1);
			if ( priv->cur_name != NULL ) {
				memcpy(priv->cur_name, name, ext_l);
				priv->cur_name[ext_l] = '\0';
			}
			return 1;	/* found, if cur_name==NULL then malloc error */
		}
		return 0;	/* found, but no delete (RENAME etc...) */
	}
	return 0;	/* not found */
}

/* get boundary string of mime multipart message */
char *get_boundary(struct mlfiPriv *priv, char *p)
{
	char ech;	/* boundary end char '"' or '\r' */
	char *rp;
	int l;

	rp = p;
	while ( *p ) {
		if ( strncasecmp(p, "boundary", 8) == 0 ) {
			p += 8;
			while ( *p && isspace(*p) ) ++p;
			if ( *p == '=' ) {
				if ( priv->multipart_level >= MAX_BOUNDARY ) return NULL;
				++p;
				while ( *p && isspace(*p) ) ++p;
				if ( *p == '"' ) {
					++p;
					ech = '"';
				} else {
					ech = '\r';
				}
				rp = p;

				l = 0;
				while ( *p && *p != '\n' && *p != ech ) {
					++l;
					++p;
				}
				p = malloc(l+1);
				if ( p == NULL ) return NULL;

				strncpy(p, rp, l);
				p[l] = '\0';

				DEBUG_fprintf((debug_fp, "BOUNDARY[%d] = %s\n", priv->multipart_level, p));

				priv->boundary[priv->multipart_level] = p;
				++priv->multipart_level;
			}
		} else {
			while ( *p && *p != ';' ) ++p;
			if ( *p == ';' ) ++p;
			while ( isspace(*p) ) ++p;
		}
	}
	return rp;
}

/* add header to own buffer */
int add_header(struct mlfiPriv *priv, char *f, char *v)
{
	struct HEADER *hd;

	hd = malloc(sizeof(*hd));
	if ( hd == NULL ) return -1;
	hd->next = NULL;
	hd->f = strdup(f);
	if ( hd->f == NULL ) return -1;
	hd->v = strdup(v);
	if ( hd->v == NULL ) return -1;
	hd->pre = priv->last_header;
	if ( priv->last_header == NULL ) {
		priv->header = hd;
	} else {
		priv->last_header->next = hd;
	}
	priv->last_header = hd;
	return 0;
}

/* release all own headers */
void clear_header(struct mlfiPriv *priv)
{
	struct HEADER *hd, *phd;

	hd = priv->last_header;
	while ( hd != NULL ) {
		phd = hd->pre;
		free(hd->f);
		free(hd->v);
		free(hd);
		hd = phd;
	}
}

/* start new mail */
sfsistat mlfi_envfrom(SMFICTX *ctx, char **envfrom)
{
	struct mlfiPriv *priv;

	DEBUG_fprintf((debug_fp, "\n*** start from %s\n", *envfrom));

	/* allocate some private memory */
	priv = malloc(sizeof *priv);
	if ( priv == NULL ) {
		/* can't accept this message right now */
		return SMFIS_TEMPFAIL;
	}
	memset(priv, '\0', sizeof *priv);
	priv->final_action = SMFIS_CONTINUE;

	/* save the private data */
	if ( !stream_mode ) {
		smfi_setpriv(ctx, priv);
	} else {
		local_priv = priv;
	}

	if ( add_header(priv, "$envelope-from", *envfrom) != 0 ) {
		return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
	}

	/* continue processing */
	return SMFIS_CONTINUE;
}

/* envelope recipiant */
sfsistat mlfi_envrcpt(SMFICTX *ctx, char **envrcpt)
{
	struct mlfiPriv *priv = MLFIPRIV;

	DEBUG_fprintf((debug_fp, "\n*** to %s\n", *envrcpt));

	if ( add_header(priv, "$envelope-to", *envrcpt) != 0 ) {
		return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
	}

	/* continue processing */
	return SMFIS_CONTINUE;
}

/* sereach filename of mail header part */
sfsistat search_header_file_name(struct mlfiPriv *priv, char *p)
{
	while ( *p ) {
		if ( strncasecmp(p, "name", 4) == 0 ) {
			p += 4;
			while ( isspace(*p) ) ++p;
			if ( *p != '=' ) break;
			++p;
			while ( isspace(*p) ) ++p;
			if ( check_bad_ext(priv, p, 0) ) {
				priv->bad_name = priv->cur_name;
				if ( priv->save == 0 && priv->final_action != SMFIS_CONTINUE ) {
					filter_log(priv);
					return priv->final_action;
				}

				if ( priv->bad_name == NULL ) {
					return SMFIS_TEMPFAIL;
				}
			}
		}
		++p;
	}
	return SMFIS_CONTINUE;
}

/* handle each header */
sfsistat mlfi_header(SMFICTX *ctx, char *headerf, char *headerv)
{
	struct mlfiPriv *priv = MLFIPRIV;
	sfsistat act;

	if ( add_header(priv, headerf, headerv) != 0 ) {
		return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
	}

	if ( strcasecmp(headerf, "Content-Type") == 0 ) {
		DEBUG_fprintf((debug_fp, "Hcontent-type %s\n", headerv));

		while ( isspace(*headerv) ) ++headerv;
		if ( strncasecmp(headerv, "multipart/", 10) == 0 ) {
			DEBUG_fprintf((debug_fp, "Hmultipart\n"));

			headerv += 10;
			if ( strncasecmp(headerv, "alternative", 11) == 0 ) {
				priv->alternative = 1;
			}
			while ( *headerv && *headerv != ';' ) ++headerv;
			if ( *headerv == ';' ) ++headerv;
			while ( *headerv && isspace(*headerv) ) ++headerv;
			if ( get_boundary(priv, headerv) == NULL ) {
				return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
			}
		} else if ( strncasecmp(headerv, "text/", 5) == 0 ) {
			if ( strncasecmp(headerv, "text/plain", 10) == 0 ) {
				priv->text_plain = strdup(headerv);
			} else if ( strncasecmp(headerv, "text/html", 9) == 0 ) {
				priv->text_html = strdup(headerv);
			}
		} else {
			act = search_header_file_name(priv, headerv);
			if ( act != SMFIS_CONTINUE ) {
				return mlfi_cleanup(ctx, act);
			}
		}
	} else if ( strcasecmp(headerf, "Content-Disposition") == 0 ) {
		act = search_header_file_name(priv, headerv);
		if ( act != SMFIS_CONTINUE ) {
			return mlfi_cleanup(ctx, act);
		}
	} else if ( strcasecmp(headerf, "Content-Transfer-Encoding") == 0 ) {
		if ( strncasecmp(headerv, "quoted-printable", 16) == 0 ) {
			if ( priv->text_html != NULL ) {
				priv->html_encode = 1;
			}
			if ( priv->text_plain != NULL ) {
				priv->text_encode = 1;
			}
		}
	}
	/* continue processing */
	return SMFIS_CONTINUE;
}

/* mail header end */
sfsistat mlfi_eoh(SMFICTX *ctx)
{
	/* continue processing */
	return SMFIS_CONTINUE;
}

/* found multipart header with filename, check filename */
sfsistat search_multi_file_name(struct mlfiPriv *priv, char *p)
{
	priv->change_extp = NULL;

	while ( *p ) {
		if ( strncasecmp(p, "name", 4) == 0 ) {
			DEBUG_fprintf((debug_fp, "Bname %s", p));

			p += 4;
			while ( isspace(*p) ) ++p;
			if ( *p == '=' ) {
				++p;
				while ( isspace(*p) ) ++p;

				/* check filename */
				if ( *p == '"' ) ++p;
				if ( check_bad_ext(priv, p, 1) ) {
					if ( priv->bad_name_type == 0 ) {
						priv->bad_name = priv->cur_name;
					} else {
						if ( priv->cur_name != NULL ) {
							free(priv->cur_name);
						}
					}
					DEBUG_fprintf((debug_fp, "** BAD extname\n"));

					if ( priv->save == 0 && priv->final_action != SMFIS_CONTINUE ) {
						filter_log(priv);
						return priv->final_action;
					}

					priv->bad_name_type = 1;
					if ( priv->bad_name == NULL ) {		/* malloc error */
						return SMFIS_TEMPFAIL;
					}

					break;
				}
			}
		} else {
			++p;
		}
	}
	return SMFIS_CONTINUE;
}

/* "begin" found, check filename */
sfsistat search_uu_file_name(struct mlfiPriv *priv, char *p)
{
	priv->change_extp = NULL;

	while ( *p && *p != '\n' && isspace(*p) ) ++p;
	if ( isdigit(p[0]) && isdigit(p[1]) && isdigit(p[2]) ) {
		/* skip file mode */
		p += 3;
		while ( *p && *p != '\n' && isspace(*p) ) ++p;
		if ( *p == '"' ) ++p;

		/* check filename */
		if ( check_bad_ext(priv, p, 1) ) {
			DEBUG_fprintf((debug_fp, "** BAD extname\n"));

			if ( priv->bad_name_type == 0 ) {
				priv->bad_name = priv->cur_name;
			} else {
				if ( priv->cur_name != NULL ) {
					free(priv->cur_name);
				}
			}

			if ( priv->save == 0 && priv->final_action != SMFIS_CONTINUE ) {
				filter_log(priv);
				return priv->final_action;
			}

			if ( priv->bad_name == NULL ) {		/* malloc error */
				return SMFIS_TEMPFAIL;
			}
			if ( priv->block_start_pos == 0 ) {
				/* uuencoded file in text part */
				priv->bad_name_type = 2;
				priv->block_start_pos = priv->body_cur_pos;
				priv->text_add_pos = priv->body_cur_pos;
			} else {
				/* uuencoded file in binary part */
				priv->bad_name_type = 1;
			}
		}
	}
	return SMFIS_CONTINUE;
}

/* Q encode the message to new allocated buffer */
char *Qencode(unsigned char *msg)
{
	char *dst;
	char *p;
	char hex[] = "0123456789ABCDEF";

	dst = malloc(strlen(msg)*3+1);	/* max 3x */
	if ( dst == NULL ) return NULL;
	p = dst;
	while ( *msg ) {
		if ( *msg == '\r' || *msg == '\n' ) {
			*p++ = *msg++;
		} else if ( *msg < 0x20 || *msg == '=' || *msg >= 0x7f ) {
			*p++ = '=';
			*p++ = hex[(*msg)>>4];
			*p++ = hex[(*msg)&0xf];
			++msg;
		} else {
			*p++ = *msg++;
		}
	}
	*p = '\0';
	return dst;
}

/* add message to mail body's text part */
int add_message(struct mlfiPriv *priv, int remain)
{
	char *from, *to;
	int l, move_len;
	int mh_len;
	char *msg;
	char *msgh;

	/* text/plain part, add msg before text end */
	if ( priv->text_add_pos != 0 ) {
		if ( priv->text_encode ) {		/* quoted-printable? */
			msg = Qencode(priv->add_msg);
			if ( msg == NULL ) return -1;
			msgh = Qencode(msg_hd);
			if ( msgh == NULL ) return -1;
		} else {
			msg = strdup(priv->add_msg);
			if ( msg == NULL ) return -1;
			msgh = strdup(msg_hd);
			if ( msgh == NULL ) return -1;
		}
		mh_len = strlen(msgh);
		l = mh_len+strlen(msg)+2+2;
		priv->body = realloc(priv->body, priv->body_len+remain+l+1);
		if ( priv->body == NULL ) {
			return -1;
		}
		from = &priv->body[priv->text_add_pos];
		to = from+l;
		move_len = priv->body_len - priv->text_add_pos +1;
		memmove(to, from, move_len);
		priv->body_len += l;
		memcpy(from, msgh, mh_len);
		memcpy(from+mh_len, msg, l-mh_len-4);
		memcpy(from+l-4, "\r\n\r\n", 4);
		priv->body[priv->body_len] = '\0';
		if ( priv->html_add_pos >= priv->text_add_pos ) {
			priv->html_add_pos += l;
		}
		if ( priv->block_start_pos >= priv->text_add_pos ) {
			priv->block_start_pos += l;
		}
		if ( priv->content_type_pos >= priv->text_add_pos ) {
			priv->content_type_pos += l;
		}
		priv->body_cur_pos += l;
		priv->text_add_pos += l;
		free(msg);
		free(msgh);
	}

	/* text/html part, add <pre>msg</pre> before </body> */
	if ( priv->html_add_pos != 0 ) {
		if ( priv->html_encode ) {		/* quoted-printable? */
			msg = Qencode(priv->add_msg);
			if ( msg == NULL ) return -1;
			msgh = Qencode(msg_hd);
			if ( msgh == NULL ) return -1;
		} else {
			msg = strdup(priv->add_msg);
			if ( msg == NULL ) return -1;
			msgh = strdup(msg_hd);
			if ( msgh == NULL ) return -1;
		}
		mh_len = strlen(msgh);
		l = mh_len+strlen(msg)+15;		/* <pre>\r\n </pre>\r\n */
		priv->body = realloc(priv->body, priv->body_len+remain+l+1);
		if ( priv->body == NULL ) {
			return -1;
		}
		from = &priv->body[priv->html_add_pos];
		to = from+l;
		move_len = priv->body_len - priv->html_add_pos +1;
		memmove(to, from, move_len);
		priv->body_len += l;
		memcpy(from, "<pre>", 5);
		memcpy(from+5, msgh, mh_len);
		memcpy(from+5+mh_len, msg, l-mh_len-15);
		memcpy(from+l-10, "\r\n</pre>\r\n", 10);
		priv->body[priv->body_len] = '\0';
		if ( priv->text_add_pos >= priv->html_add_pos ) {
			priv->text_add_pos += l;
		}
		if ( priv->block_start_pos >= priv->html_add_pos ) {
			priv->block_start_pos += l;
		}
		if ( priv->content_type_pos >= priv->html_add_pos ) {
			priv->content_type_pos += l;
		}
		priv->body_cur_pos += l;
		priv->html_add_pos += l;
		free(msg);
		free(msgh);
	}
	priv->add_msg = NULL;
	return 0;
}

/* change content type to application/octet-stream for RENAME */
int change_content_type(struct mlfiPriv *priv, int remain_len)
{
	char *from, *to;
	int old_len, new_len, move_len;

	if ( priv->content_type_pos == 0 ) return 0;
	from = &priv->body[priv->content_type_pos];
	old_len = 0;
	while ( *from && !isspace(*from) && *from != ';' ) {
		++from;
		++old_len;
	}
	new_len = 24;			/* strlen("application/octet-stream") */
	if ( old_len < new_len ) {
		priv->body = realloc(priv->body, priv->body_len+new_len-old_len+remain_len+1);
		if ( priv->body == NULL ) return -1;
	}

	if ( old_len != new_len ) {
		from = &priv->body[priv->content_type_pos];
		from += old_len;
		to = from-old_len+new_len;
		move_len = priv->body_len - (priv->content_type_pos + old_len) +1;
		memmove(to, from, move_len);
	}

	from = &priv->body[priv->content_type_pos];
	memcpy(from, "application/octet-stream", 24);

/*
	** content_type is always last block, no need customise positions

	if ( priv->text_add_pos >= priv->content_type_pos ) {
		priv->text_add_pos += new_len - old_len;
	}
	if ( priv->html_add_pos >= priv->content_type_pos ) {
		priv->html_add_pos += new_len - old_len;
	}
	if ( priv->block_start_pos >= priv->content_type_pos ) {
		priv->block_start_pos += new_len - old_len;
	}
*/
	return 0;
}

/* handle body part */
sfsistat mlfi_body(SMFICTX *ctx, u_char *bodyp, size_t bodylen)
{
	struct mlfiPriv *priv = MLFIPRIV;
	char *p;
	int l;
	int multi_end;	/* multipart boundary found, next/end block */
	sfsistat act;
	char uc;

	if ( priv->save != 0 ) {
		if ( backup_mail(priv, bodyp, bodylen) != 0 ) {
			return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
		}
	}

	/* allocate body by new size */
	if ( priv->body == NULL ) {
		priv->body = malloc(bodylen+1);
	} else {
		priv->body = realloc(priv->body, priv->body_len+bodylen+1);
	}
	if ( priv->body == NULL ) {
		return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
	}

	while ( bodylen > 0 ) {

		/* get one line */
		p = &priv->body[priv->body_len];
		while ( bodylen-- > 0 ) {
			++priv->body_len;
			if ( *bodyp == '\n' ) {		/* end of line */
				*p++ = *bodyp++;
				*p = '\0';
				break;
			}
			*p++ = *bodyp++;
			if ( bodylen == 0 ) {
				return SMFIS_CONTINUE;	/* not line end */
			}
		}

		/* check this line or one header */
		p = &priv->body[priv->body_cur_pos];
		if ( priv->now_multipart_header ) {
			/* handle header of multipart */
			if ( *p == ' ' || *p == '\t' ) {
				/* continue header by folding, stay positon, get next line */
				priv->body_cur_pos = priv->body_len;
				continue;
			}
			uc = *p;			/* prepare next header top char */
			*p = '\0';
			p = &priv->body[priv->multi_header_pos];

			DEBUG_fprintf((debug_fp, "Bm-head %s", p));

			if ( strncasecmp(p, "Content-Type:", 13) == 0 ) {
				DEBUG_fprintf((debug_fp, "Bcontent-type %s", p));

				p += 13;
				while ( isspace(*p) ) ++p;
				priv->content_type_pos = p-priv->body;		/* mark type for RENAME */
				if ( strncasecmp(p, "multipart/", 10) == 0 ) {
					/* multipart nesting */
					DEBUG_fprintf((debug_fp, "Bmultipart, search-boundary %s", p));

					p += 10;
					if ( strncasecmp(p, "alternative", 11) == 0 ) {
						priv->alternative = 1;
					}
					p = get_boundary(priv, p);
					if ( p == NULL ) {
						return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
					}
				} else if ( strncasecmp(p, "text/plain", 10) == 0 ) {
					if ( priv->text_plain == NULL ) {
						DEBUG_fprintf((debug_fp, "Btext/plain %s", p));
						priv->text_plain = strdup(p);
						if ( priv->text_plain == NULL ) {
							return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
						}
						if ( priv->alternative == 1 || priv->html_add_pos == 0 ) {
							priv->multi_body_type = 1;
						}
					}
				} else if ( strncasecmp(p, "text/html", 9) == 0 ) {
					if ( priv->text_html == NULL ) {
						DEBUG_fprintf((debug_fp, "Btext/html %s", p));
						priv->text_html = strdup(p);
						if ( priv->text_html == NULL ) {
							return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
						}
						if ( priv->alternative == 1 || priv->text_add_pos == 0 ) {
							priv->multi_body_type = 2;
						}
					}
				} else if ( strncasecmp(p, "message/rfc822", 14) == 0 ) {
					priv->next_body_is_header = 1;
				} else {
					DEBUG_fprintf((debug_fp, "Bsearch filename %s", p));
					act = search_multi_file_name(priv, p);
					if ( act != SMFIS_CONTINUE ) {
						return mlfi_cleanup(ctx, act);
					}
				}
			} else if ( strncasecmp(p, "Content-Disposition:", 20) == 0 ) {
				DEBUG_fprintf((debug_fp, "Content-Disposition %s", p));
				p += 20;
				while ( isspace(*p) ) ++p;

				act = search_multi_file_name(priv, p);
				if ( act != SMFIS_CONTINUE ) {
					return mlfi_cleanup(ctx, act);
				}
			} else if ( strncasecmp(p, "Content-Transfer-Encoding:", 26) == 0 ) {
				p += 26;
				while ( isspace(*p) ) ++p;
				if ( strncasecmp(p, "quoted-printable", 16) == 0 ) {
					if ( priv->multi_body_type == 2 ) {
						priv->html_encode = 1;
					}
					if ( priv->multi_body_type == 1 ) {
						priv->text_encode = 1;
					}
				}
			}

			priv->multi_header_pos = priv->body_cur_pos;	/* for next header */

			p = &priv->body[priv->body_cur_pos];	/* line start pos */
			*p = uc;			/* restore next header top char */

			if ( strcmp(p, "\r\n") == 0 || strcmp(p, "\n") == 0 ) {
				DEBUG_fprintf((debug_fp, "Bmulti header end\n"));

				priv->now_multipart_header = 0;
				if ( priv->next_body_is_header ) {
					priv->next_body_is_header = 0;
					priv->now_multipart_header = 1;
				}
			}

			if ( priv->save == 1 ) {
				if ( backup_mail(priv, priv->body, priv->body_len) != 0 ) {
					return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
				}
				if ( backup_mail(priv, bodyp, bodylen) != 0 ) {
					return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
				}
			}

			if ( priv->change_extp != NULL ) {
				if ( *priv->change_extp == '.' ) {
					*priv->change_extp = '_';
				}
				priv->change_extp = NULL;
				if ( change_content_type(priv, bodylen) != 0 ) {
					return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
				}
			}

			if ( priv->add_msg != NULL ) {
				if ( add_message(priv, bodylen) != 0 ) {
					return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
				}
			}

		} else {
/*
			DEBUG_fprintf((debug_fp, "body[%d]\n", priv->multipart_level));
*/
			if ( strncmp(p, "begin ", 6) == 0 ) {
				DEBUG_fprintf((debug_fp, "Buuencode-begin %s", p));
				p += 6;

				act = search_uu_file_name(priv, p);
				if ( act != SMFIS_CONTINUE ) {
					return mlfi_cleanup(ctx, act);
				}
				if ( priv->save == 1 ) {
					if ( backup_mail(priv, priv->body, priv->body_len) != 0 ) {
						return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
					}
					if ( backup_mail(priv, bodyp, bodylen) != 0 ) {
						return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
					}
				}

				if ( priv->change_extp != NULL ) {
					if ( *priv->change_extp == '.' ) {
						*priv->change_extp = '_';
					}
					priv->change_extp = NULL;
					if ( change_content_type(priv, bodylen) != 0 ) {
						return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
					}
				}
				if ( priv->add_msg != NULL ) {
					if ( add_message(priv, bodylen) != 0 ) {
						return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
					}
				}
			} else if ( priv->bad_name_type == 2 && strncmp(p, "end", 3) == 0 ) {
				/* search uuencode 'end' */
				/* calc new body size of deleted binary */
				l = priv->block_start_pos
					+ strlen(uu_cut[0])
					+ strlen(priv->bad_name)
					+ strlen(uu_cut[1]);
				priv->body = realloc(priv->body, l+bodylen+1);
				if ( priv->body == NULL ) {
					return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
				}

				/* add message replace of binary */
				p = &priv->body[priv->block_start_pos];
				memcpy(p, uu_cut[0], strlen(uu_cut[0]));
				p += strlen(uu_cut[0]);
				memcpy(p, priv->bad_name, strlen(priv->bad_name));
				p += strlen(priv->bad_name);
				memcpy(p, uu_cut[1], strlen(uu_cut[1]));
				p += strlen(uu_cut[1]);
				*p = '\0';
				free(priv->bad_name);
				priv->bad_name = NULL;
				priv->bad_name_type = 0;
				priv->block_start_pos = 0;
				priv->body_len = l;
				priv->body_cur_pos = l;
				l = 0;
				priv->body_changed = 1;
				priv->msg_added = 0;
			} else if ( priv->multipart_level != 0 && p[0] == '-' && p[1] == '-' ) {
				/* search multipart block boundary */
				DEBUG_fprintf((debug_fp, "Bmulti next check %s", p));

				p += 2;
				if ( strncmp(p, priv->boundary[priv->multipart_level-1], strlen(priv->boundary[priv->multipart_level-1])) == 0 ) {
					p += strlen(priv->boundary[priv->multipart_level-1]);
					if ( p[0] == '-' && p[1] == '-' ) {
						multi_end = 1;
					} else {
						multi_end = 0;
					}

					if ( priv->multi_body_type == 1 ) {
						/* text message adding position, as text/plain block end */
						if ( priv->text_add_pos == 0 ) {
							priv->text_add_pos = priv->body_cur_pos;
						}
					}

					if ( priv->bad_name_type == 1 ) {
						/* change block to mes */
						l = priv->block_start_pos
							+ strlen(mime_cut[0])
							+ strlen(priv->bad_name)
							+ strlen(mime_cut[1])
							+ strlen(priv->boundary[priv->multipart_level-1])
							+ strlen(mime_cut[2]);
						if ( multi_end ) l += 2;
						priv->body = realloc(priv->body, l+bodylen+1);
						if ( priv->body == NULL ) {
							return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
						}
						p = &priv->body[priv->block_start_pos];
						memcpy(p, mime_cut[0], strlen(mime_cut[0]));
						p += strlen(mime_cut[0]);
						memcpy(p, priv->bad_name, strlen(priv->bad_name));
						p += strlen(priv->bad_name);
						memcpy(p, mime_cut[1], strlen(mime_cut[1]));
						p += strlen(mime_cut[1]);
						memcpy(p, priv->boundary[priv->multipart_level-1], strlen(priv->boundary[priv->multipart_level-1]));
						p += strlen(priv->boundary[priv->multipart_level-1]);
						if ( multi_end ) {
							*p++ = '-';
							*p++ = '-';
						}
						memcpy(p, mime_cut[2], strlen(mime_cut[2]));
						p += strlen(mime_cut[2]);
						*p = '\0';
						free(priv->bad_name);
						priv->bad_name = NULL;
						priv->bad_name_type = 0;
						priv->block_start_pos = 0;
						priv->body_len = l;
						priv->body_cur_pos = l;
						l = 0;
						priv->body_changed = 1;
					}

					if ( multi_end ) {
						DEBUG_fprintf((debug_fp, "Bmulti end\n"));

						priv->block_start_pos = 0;
						--priv->multipart_level;
						free(priv->boundary[priv->multipart_level]);
						priv->boundary[priv->multipart_level] = NULL;
					} else {
						DEBUG_fprintf((debug_fp, "Bmulti start\n"));

						priv->now_multipart_header = 1;
						priv->block_start_pos = priv->body_len;
						priv->multi_header_pos = priv->body_len;
					}
					priv->msg_added = 0;
					priv->content_type_pos = 0;
					priv->multi_body_type = 0;
				}
			} else if ( priv->multi_body_type == 2 ) {
				/* search html message adding position, as before </body> */
				if ( priv->html_add_pos == 0 ) {
					l = str_in_buf(p, "</body");
					if ( l != -1 ) {
						priv->html_add_pos = priv->body_cur_pos+l;
					}
				}
			}
		}
		priv->body_cur_pos = priv->body_len;
	}

	/* continue processing */
	return SMFIS_CONTINUE;
}

/* end of mail */
sfsistat mlfi_eom(SMFICTX *ctx)
{
	struct mlfiPriv *priv = MLFIPRIV;
	char *hd;

	if ( priv->final_action != SMFIS_CONTINUE ) {
		filter_log(priv);
		return mlfi_cleanup(ctx, priv->final_action);
	}

	if ( priv->body_changed ) {
		if ( id_str != NULL ) {
			hd = malloc(strlen(release_version)+strlen(id_str)+4+1);
			if ( hd == NULL ) {
				return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
			}
			strcpy(hd,release_version);
			if ( stream_mode ) {
				strcat(hd,";\n\t");
			} else {
				strcat(hd,";\r\n\t");
			}
			strcat(hd,id_str);
			if ( !stream_mode ) {
				smfi_addheader(ctx, "X-Mail-Filter", hd);
			} else {
				stream_addheader("X-Mail-Filter", hd);
			}
			free(hd);
		} else {
			if ( !stream_mode ) {
				smfi_addheader(ctx, "X-Mail-Filter", release_version);
			} else {
				stream_addheader("X-Mail-Filter", release_version);
			}
		}
		if ( priv->save != 0 ) {
			hd = malloc(strlen(priv->savename)+2+1);
			if ( hd == NULL ) {
				return mlfi_cleanup(ctx, SMFIS_TEMPFAIL);
			}
			hd[0] = '"';
			strcpy(hd+1,priv->savename);
			strcat(hd,"\"");
			if ( !stream_mode ) {
				smfi_addheader(ctx, "X-Backup-File", hd);
			} else {
				stream_addheader("X-Backup-File", hd);
			}
		}
		if ( !stream_mode ) {
			smfi_replacebody(ctx, priv->body, priv->body_len);
		} else {
			stream_replacebody(priv->body, priv->body_len);
		}
	}
	return mlfi_cleanup(ctx, SMFIS_CONTINUE);
}

sfsistat mlfi_close(SMFICTX *ctx)
{
	return SMFIS_ACCEPT;
}

/* abort while mail recieving */
sfsistat mlfi_abort(SMFICTX *ctx)
{
	return mlfi_cleanup(ctx, SMFIS_CONTINUE);
}

/* clear working data for end/abort mail handling */
sfsistat mlfi_cleanup(SMFICTX *ctx, sfsistat act)
{
	struct mlfiPriv *priv = MLFIPRIV;
	int i;

#ifdef DEBUG
	fflush(debug_fp);
#endif
	if ( priv == NULL ) {
		return SMFIS_CONTINUE;
	}

	/* set reply for REJECT */
	if ( act == SMFIS_REJECT ) {
		set_reject_reply(ctx);
	}

	for ( i = 0; i < MAX_BOUNDARY; ++i ) {
		if ( priv->boundary[i] != NULL ) {
			free(priv->boundary[i]);
			priv->boundary[i] = NULL;
		}
	}

	if ( priv->body != NULL ) free(priv->body);
	if ( priv->bad_name != NULL ) free(priv->bad_name);
	if ( priv->text_plain != NULL ) free(priv->text_plain);
	if ( priv->text_html != NULL ) free(priv->text_html);
	if ( priv->savename != NULL ) free(priv->savename);
	if ( priv->savefp != NULL ) fclose(priv->savefp);

	clear_header(priv);

	/* release private memory */
	free(priv);
	if ( !stream_mode ) {
		smfi_setpriv(ctx, NULL);
	}

	/* return status */
	return act;
}

/* display rule file error */
void error(char *err_message, char *bad_string)
{
	fprintf(stderr, "%s:%d %s", fname, read_lines, err_message);
	if ( bad_string != NULL ) {
		fprintf(stderr, " : ");
		while ( *bad_string ) {
			if ( *bad_string == '\n' ) break;
			fputc(*bad_string, stderr);
			++bad_string;
		}
	}
	fprintf(stderr, "\n");
	exit(EX_USAGE);
}

/* add one rule to rules */
int add_rule(struct CHECK_RULE *rule)
{
	struct CHECK_RULE *rp;

	if ( rule->action == NULL ) {
		error("ACTION not found", NULL);
	}
	if ( rule->cond == NULL ) {
		rule->cond = "";
	}
	rule->rule_n = read_lines-1;
	rp = &rules;
	while ( rp->next != NULL ) {
		rp = rp->next;
	}
	rp->next = malloc(sizeof(*rule));
	if ( rp->next == NULL ) {
		return -1;
	}
	memcpy(rp->next, rule, sizeof(*rule));
	if ( verbose ) {
		output_rule(stdout, rule);
	}
	return 0;
}

char *skip_comma(char *p)
{
	while ( isspace(*p) ) ++p;
	if ( *p == ',' ) ++p;
	while ( isspace(*p) ) ++p;
	return p;
}

/* cutting one word from buf to word, break if termchars found */
/* return new buf pos */
char *get_word(char *buf, char *word, char *termchars)
{
	char *q;

	while ( isspace(*buf) ) ++buf;
	if ( *buf == '"' ) {
		++buf;
		while ( *buf && *buf != '"' ) {
			*word++ = *buf++;
		}
		*word = '\0';
		if ( *buf != '"' ) error("syntax error", buf);
		++buf;
	} else {
		while ( *buf && !isspace(*buf) ) {
			q = termchars;
			while ( *q ) {
				if ( *buf == *q ) {
					*word = '\0';
					return buf;
				}
				++q;
			}
			*word++ = *buf++;
		}
		*word = '\0';
	}
	while ( isspace(*buf) ) ++buf;
	return buf;
}

/* read rule from file */
void read_rules(char *filename)
{
	FILE *fp;
	char buf[1024];
	char wk[1024];
	struct CHECK_RULE rule;
	char *p, *q;
	int reading;	/* 0:ext 1:cond 2:cond string 3:action 4:"MSG" */
	int added_len;	/* previous length of added string */
	int l;
	int save_exist = 0;

	fname = filename;
	fp = fopen(filename, "r");
	if ( fp == NULL ) {
		perror(filename);
		exit(EX_USAGE);
	}
	read_lines = 0;
	memset(&rule, 0, sizeof(rule));
	added_len = 0;
	reading = 0;
	if ( verbose ) printf("readed rule:\n");
	while ( fgets(buf, 1024, fp) != NULL ) {
		++read_lines;
		p = buf;
		if ( reading != 5 ) {				/* newline in "" ? */
			if ( *p == '#' ) continue;
			if ( *p != ' ' && *p != '\t' ) {	/* new rule? */
				if ( rule.ext != NULL ) {
					if ( reading == 4 ) {
						error("bad MSG string", NULL);
					}
					add_rule(&rule);
					memset(&rule, 0, sizeof(rule));
					added_len = 0;
					reading = 0;
				}
			}
			while ( isspace(*p) ) ++p;
		} else if ( reading == 5 && *p == '\t' ) {	/* skip MSG's TAB after newline */
			while ( *p == '\t' ) ++p;
		}

		/* handle word by words */
		while ( *p ) {
			if ( reading != 5 && *p == '#' ) break;
			switch ( reading ) {
			case 0:		/* ext */
				p = skip_comma(p);
				if ( *p == '\0' ) break;
				if ( *p == '?' ) {				/* ext end, next is cond */
					reading = 1;
					added_len = 0;
					if ( rule.ext == NULL ) error("syntax error", p);
					++p;
					break;
				} else if ( *p == ':' ) {		/* ext end, next is action */
					reading = 3;
					added_len = 0;
					if ( rule.ext == NULL ) error("syntax error", p);
					++p;
					break;
				}
				p = get_word(p, wk, ",:?=");
				if ( *p == '=' ) {
					++p;
					if ( strcasecmp(wk, "SAVEDIR") == 0 ) {
						p = get_word(p, wk, "");
						l = strlen(wk);
						if ( l == 0 ) {
							strcpy(wk,"./");
						} else if ( wk[l-1] != '/' ) {
							strcat(wk,"/");
						}
						save_dir = strdup(wk);
					} else if ( strcasecmp(wk, "SAVEEXPIRE") == 0 ) {
						p = get_word(p, wk, "");
						save_expire = atoi(wk);
					} else if ( strcasecmp(wk, "ID_STR") == 0 ) {
						p = get_word(p, wk, "");
						id_str = strdup(wk);
					} else {
						error("unknown variable", wk);
					}
					if ( *p && *p != '#' ) {
						error("syntax error", p);
					}
					break;
				}
				q = wk;
				if ( *q == '.' ) {
					++q;
					if ( *q == '\0' ) error("bad extension", ".");
				}
				if ( added_len == 0 ) {
					rule.ext = malloc(strlen(q)+1+1);
				} else {
					rule.ext = realloc(rule.ext, added_len+strlen(q)+1+1);
				}
				if ( rule.ext == NULL ) error("less memory", NULL);
				strcpy(&rule.ext[added_len], q);
				added_len += strlen(q)+1;
				rule.ext[added_len] = '\0';	/* for terminate */
				break;

			case 1:		/* condition */
				if ( *p == '&' ) {			/* cond string end, next is cond */
					reading = 1;
					++p;
					break;
				} else if ( *p == ':' ) {		/* cond string end, next is action */
					reading = 3;
					added_len = 0;
					++p;
					break;
				}
				p = get_word(p, wk, ":");
				if ( added_len == 0 ) {
					rule.cond = malloc(strlen(wk)+1+1);
				} else {
					rule.cond = realloc(rule.cond, added_len+strlen(wk)+1+1);
				}
				if ( rule.cond == NULL ) error("less memory", NULL);
				strcpy(&rule.cond[added_len], wk);
				added_len += strlen(wk)+1;
				rule.cond[added_len] = '\0';	/* for terminate */
				if ( *p != ':' ) error("header word need ':'", p);
				q = wk;
				if ( *q == '!' ) ++q;
				if ( *q == '$' ) {
					if ( strcasecmp(q, "$envelope-from") != 0
						&& strcasecmp(q, "$envelope-to") != 0
						&& strcasecmp(q, "$texttype") != 0
						&& strcasecmp(q, "$content-type") != 0
						&& strcasecmp(q, "$preext") != 0 ) {
						error("unknown special header", q);
					}
				}
				reading = 2;			/* cond word end, next is cond string */
				++p;					/* skip : */
				break;

			case 2:		/* condition's string */
				p = get_word(p, wk, "&:");
				rule.cond = realloc(rule.cond, added_len+strlen(wk)+1+1);
				if ( rule.cond == NULL ) error("less memory", NULL);
				strcpy(&rule.cond[added_len], wk);
				added_len += strlen(wk)+1;
				rule.cond[added_len] = '\0';	/* for terminate */
				reading = 1;
				break;

			case 3:		/* action */
				p = skip_comma(p);
				if ( *p == '\0' ) break;
				p = get_word(p, wk, ",");
				if ( added_len == 0 ) {
					rule.action = malloc(strlen(wk)+1+1);
				} else {
					rule.action = realloc(rule.action, added_len+strlen(wk)+1+1);
				}
				if ( rule.action == NULL ) error("less memory", NULL);
				strcpy(&rule.action[added_len], wk);
				added_len += strlen(wk)+1;
				rule.action[added_len] = '\0';	/* for terminate */
				if ( strcasecmp(wk, "SAVE") == 0 ) {
					save_exist = 1;
				}
				if ( strcasecmp(wk, "MSG") == 0 ) {
					if ( rule.ext[0] == '-' ) {
						error("- part can't use", wk);
					}
					reading = 4;
				} else if ( strcasecmp(wk, "OK") != 0
					&& strcasecmp(wk, "REJECT") != 0
					&& strcasecmp(wk, "DROP") != 0
					&& strcasecmp(wk, "DELETE") != 0
					&& strcasecmp(wk, "RENAME") != 0
					&& strcasecmp(wk, "SAVE") != 0 ) {
					error("unknown action", wk);
				} else if ( rule.ext[0] == '-' ) {
					if ( strcasecmp(wk, "DELETE") == 0
						|| strcasecmp(wk, "RENAME") == 0 ) {
						error("- part can't use", wk);
					}
				}
				break;

			case 4:		/* MSG's string */
				if ( *p == ',' ) {
					error("MSG need string", p);
				}
				if ( *p == '"' ) {
					++p;
					reading = 5;
					++added_len;
					break;
				} else {
					error("MSG need string", p);
				}
				break;

			case 5:		/* MSG's string with "" */
				q = wk;
remain_string:
				while ( *p && *p != '"' ) {
					if ( *p == '\n' ) {
						*q++ = '\r';
					}
					*q++ = *p++;
				}
				if ( *p == '"' && p[1] == '"' ) {
					*q++ = '"';
					p += 2;
/**************************************************/
/* """ -> " for JIS string                        */
/**************************************************/
					if ( *p == '"' ) ++p;

					goto remain_string;
				}
				*q = '\0';
				--added_len;
				rule.action = realloc(rule.action, added_len+strlen(wk)+1+1);
				if ( rule.action == NULL ) error("less memory", NULL);
				strcpy(&rule.action[added_len], wk);
				added_len += strlen(wk)+1;
				rule.action[added_len] = '\0';	/* for terminate */
				if ( *p == '"' ) {
					++p;
					reading = 3;
				}
				break;
			}
			while ( isspace(*p) ) ++p;
		}
	}
	fclose(fp);

	/* add last rule */
	if ( rule.ext != NULL ) {
		if ( reading == 4 || reading == 5 ) {
			error("bad MSG string", NULL);
		}
		add_rule(&rule);
	}
	if ( verbose ) printf("\n");
	if ( save_exist ) {
		if ( save_dir == NULL ) {
			error("SAVEDIR= not found", NULL);
		}
		p = malloc(strlen(save_dir)+16);
		if ( p == NULL ) {
			error("less memory", NULL);
		}
		strcpy(p, save_dir);
		strcat(p, ".extest");
		fp = fopen(p, "w");
		if ( fp == NULL ) {
			error("SAVEDIR can't write", save_dir);
		}
		fclose(fp);
		unlink(p);
	}
}

/* milter definition table for sendmail */
struct smfiDesc smfilter = {
	"AttachFilter",	/* filter name */
	SMFI_VERSION,	/* version code -- do not change */
	SMFIF_ADDHDRS|SMFIF_CHGBODY,	/* flags */
	NULL,		/* connection info filter */
	NULL,		/* SMTP HELO command filter */
	mlfi_envfrom,	/* envelope sender filter */
	mlfi_envrcpt,	/* envelope recipient filter */
	mlfi_header,	/* header filter */
	mlfi_eoh,	/* end of header */
	mlfi_body,	/* body block filter */
	mlfi_eom,	/* end of message */
	mlfi_abort,	/* message aborted */
	mlfi_close	/* connection cleanup */
};


/*
** stream filter part
** read file and send milter functions.
** not correctly simulate sendmail's behavor.
*/

char *s_headers;	/* readed headers. Header\0Headerbody\0...\0\0 */
int s_header_len;	/* s_headers size */
char *s_body;		/* readed body */
int s_body_len;		/* readed body length */

/* replace body from filter */
void stream_replacebody(char *body, int body_len)
{
	s_body_len = body_len;
	s_body = realloc(s_body, s_body_len);
	if ( s_body == NULL ) error("less memory", NULL);
	memcpy(s_body, body, s_body_len);
}

/* add header from filter */
void stream_addheader(char *f, char *v)
{
	int l;
	l = strlen(f)+1+strlen(v)+1;
	s_headers = realloc(s_headers, s_header_len+l+1);
	if ( s_headers == NULL ) error("less memory", NULL);
	strcpy(&s_headers[s_header_len], f);
	s_header_len += strlen(f)+1;
	strcpy(&s_headers[s_header_len], v);
	s_header_len += strlen(v)+1;
	s_headers[s_header_len] = '\0';
}

/* filtering one file (one mail or multi mail with unix mailbox format */
void file_filter(FILE *fp)
{
	char buf[4096+1];
	char *p, *q;
	int nl;
	int stat;
	int l;
	char *envfrom;
	char *envrcpt;

	s_headers = malloc(1);
	if ( s_headers == NULL ) error("less memory",NULL);
	s_body = malloc(1);
	if ( s_body == NULL ) error("less memory",NULL);
	fgets(buf, 4096, fp);
	do {
		stat = SMFIS_CONTINUE;
		envfrom = NULL;
		envrcpt = NULL;
		s_header_len = 0;
		s_body_len = 0;

		/* handle header */
		nl = 1;
		do {
			++read_lines;
			l = strlen(buf);
			if ( nl && (strcmp(buf,"\r\n") == 0 || strcmp(buf,"\n") == 0) ) break;
			if ( l > 0 && buf[l-1] == '\n' ) {
				nl = 1;
			} else {
				nl = 0;
			}
			if ( l > 1 && buf[l-2] == '\r' && buf[l-1] == '\n' ) {
				buf[l-2] = '\0';
				l -= 2;
			} else if ( l > 0 && buf[l-1] == '\n' ) {
				buf[l-1] = '\0';
				--l;
			}
			if ( isspace(buf[0]) ) {
				s_headers = realloc(s_headers, s_header_len+l+1+1);
				if ( s_headers == NULL ) error("less memory", NULL);
				--s_header_len;
				s_headers[s_header_len++] = '\n';
				strcpy(&s_headers[s_header_len], buf);
				s_header_len += l+1;
				s_headers[s_header_len] = '\0';	/* for terminate */
			} else {
				s_headers = realloc(s_headers, s_header_len+l+1+1+1);
				if ( s_headers == NULL ) error("less memory", NULL);
				strcpy(&s_headers[s_header_len], buf);
				p = &s_headers[s_header_len];
				if ( strncmp(p,"From ",5) == 0 ) {
					++l;
					p[l] = '\0';
					if ( p[5] != '-' ) {		/* for almail mbox */
						p += 5;
						while ( isspace(*p) ) ++p;
						envfrom = strdup(p);
					}
				} else {
					while ( *p && *p != ':' ) ++p;
					if ( *p == ':' ) {
						*p = '\0';
					} else {
						++l;
						p[l] = '\0';
					}
					p = &s_headers[s_header_len];
					if ( envfrom == NULL && strcasecmp(p, "From") == 0 ) {
						p += 5;
						while ( isspace(*p) ) ++p;
						envfrom = strdup(p);
					} else if ( envrcpt == NULL && strcasecmp(p, "To") == 0 ) {
						p += 3;
						while ( isspace(*p) ) ++p;
						envrcpt = strdup(p);
					} else if ( envfrom == NULL && strcasecmp(p, "Return-Path") == 0 ) {
						p += 12;
						while ( isspace(*p) ) ++p;
						envfrom = strdup(p);
					}
				}
				s_header_len += l+1;
				s_headers[s_header_len] = '\0';	/* for terminate */
			}
		} while ( fgets(buf, 4096, fp) != NULL );

		if ( feof(fp) ) break;

		/* header end, send headers to filter */
		if ( envfrom == NULL ) envfrom = strdup("<>");
		if ( envfrom == NULL ) error("less memory", NULL);
		if ( envrcpt == NULL ) envrcpt = strdup("<>");
		if ( envrcpt == NULL ) error("less memory", NULL);
		stat = mlfi_envfrom(NULL, &envfrom);
		if ( stat == SMFIS_CONTINUE ) {
			stat = mlfi_envrcpt(NULL, &envrcpt);
		}
		p = s_headers;
		while ( *p ) {
			if ( strncmp(p,"From ",5) == 0 ) {
				p = p+strlen(p)+1+1;
				continue;
			}
			if ( stat != SMFIS_CONTINUE ) break;
			q = p+strlen(p)+1;
			stat = mlfi_header(NULL, p, q);
			p = q+strlen(q)+1;
		}
		if ( stat == SMFIS_CONTINUE ) stat = mlfi_eoh(NULL);

		/* handle body */
		while ( fgets(buf, 4096, fp) != NULL ) {
			++read_lines;
			if ( strncmp(buf, "From ", 5) == 0 ) break;		/* next mail */
			l = strlen(buf);
			if ( l > 1 && buf[l-2] != '\r' && buf[l-1] == '\n' ) {
				buf[l-1] = '\r';
				buf[l] = '\n';
				buf[l+1] = '\0';
				++l;
			} else if ( l == 1 && buf[0] == '\n' ) {
				strcpy(buf, "\r\n");
				l = 2;
			}

			s_body = realloc(s_body, s_body_len+l);
			if ( s_body == NULL ) error("less memory", NULL);
			memcpy(&s_body[s_body_len], buf, l);
			s_body_len += l;

			/* send filter, only one line. performance is bad ;_; */
			if ( stat == SMFIS_CONTINUE ) {
				stat = mlfi_body(NULL, buf, l);
			}
		}
		if ( stat == SMFIS_CONTINUE ) stat = mlfi_eom(NULL);
		if ( stat == SMFIS_CONTINUE ) stat = mlfi_close(NULL);

		/* output mail */
		if ( stat == SMFIS_ACCEPT ) {
			p = s_headers;
			while ( *p ) {
				l = strlen(p);
				if ( p[l+1] != '\0' ) {
					printf("%s: ",p);
					p += l+1;
					l = strlen(p);
					if ( l > 1 && p[l-2] == '\r' && p[l-1] == '\n' ) {
						p[l-2] = '\0';
					} else if ( l > 0 && p[l-1] == '\n' ) {
						p[l-1] = '\0';
					}
					if ( *p == ' ' ) {
						printf("%s\n",p+1);
					} else {
						printf("%s\n",p);
					}
					p += l+1;
				} else {
					if ( l > 1 && p[l-2] == '\r' && p[l-1] == '\n' ) {
						p[l-2] = '\n';
						p[l-1] = '\0';
					}
					printf("%s\n",p);
					p += l+1+1;
				}
			}
			printf("\n");
			p = s_body;
			q = s_body;
			l = 0;
			while ( s_body_len-- > 0 ) {
				if ( s_body_len > 0 && *p == '\r' && p[1] == '\n' ) {
					++p;
				} else {
					*q++ = *p++;
					++l;
				}
			}
			fwrite(s_body,1,l,stdout);
		} else if ( test_mode ) {
			/* output REJECT/DROP mail */
			/* From,To,Subject,Date and body is REJECT/DISCARD only */
			p = s_headers;
			while ( *p ) {
				l = strlen(p);
				if ( p[l+1] != '\0' ) {
					if ( strcasecmp(p,"from") == 0 || strcasecmp(p,"date") == 0 || strcasecmp(p,"to") == 0 || strcasecmp(p,"subject") == 0 ) {
						printf("%s: ",p);
						p += l+1;
						l = strlen(p);
						if ( l > 1 && p[l-2] == '\r' && p[l-1] == '\n' ) {
							p[l-2] = '\0';
						} else if ( l > 0 && p[l-1] == '\n' ) {
							p[l-1] = '\0';
						}
						if ( *p == ' ' ) {
							printf("%s\n",p+1);
						} else {
							printf("%s\n",p);
						}
						p += l+1;
					} else {
						p += l+1;
						l = strlen(p);
						p += l+1;
					}
				} else {
					if ( l > 1 && p[l-2] == '\r' && p[l-1] == '\n' ) {
						p[l-2] = '\n';
						p[l-1] = '\0';
					}
					printf("%s\n",p);
					p += l+1+1;
				}
			}
			printf("\n");
			if ( stat == SMFIS_REJECT ) {
				printf("REJECT by exefilter\n\n");
			} else if ( stat == SMFIS_DISCARD ) {
				printf("DISCARD by exefilter\n\n");
			} else if ( stat == SMFIS_TEMPFAIL ) {
				printf("TEMPFAIL exefilter\n\n");
			}
		}
		if ( envfrom != NULL ) free(envfrom);
		if ( envrcpt != NULL ) free(envrcpt);
	} while ( !feof(fp) );
	free(s_headers);
	free(s_body);
}

int main(int argc, char *argv[])
{
	int c;
	const char *args = "p:c:vt";
	FILE *fp;

	memset(&rules, 0, sizeof(rules));
	stream_mode = 1;

	/* Process command line options */
	while ( (c = getopt(argc, argv, args)) != -1 ) {
		switch ( c ) {
		case 'p':
			/* socket name, do milter action */
			if ( optarg == NULL || *optarg == '\0' ) {
				(void) fprintf(stderr, "Illegal conn: %s\n",
					       optarg);
				exit(EX_USAGE);
			}
			(void) smfi_setconn(optarg);
			stream_mode = 0;
			break;
		case 'c':
			/* change rule file from default */
			if ( optarg == NULL || *optarg == '\0' ) {
				(void) fprintf(stderr, "Illegal rule: %s\n",
					       optarg);
				exit(EX_USAGE);
			}
			read_rules(optarg);
			break;
		case 'v':
			/* display readed rule */
			verbose = 1;
			break;
		case 't':
			/* stream mode with REJECT/DROP check */
			test_mode = 1;
			break;
		}
	}

	if ( rules.next == NULL ) {
		read_rules(DEFAULT_RULE_FILE);
	}

	if ( stream_mode == 0 ) {
		/* do milter action */

		openlog("exefilter", LOG_PID, LOG_MAIL);

		if ( smfi_register(smfilter) == MI_FAILURE ) {
			syslog(LOG_ERR, "smfi_register failed\n");
			fprintf(stderr, "smfi_register failed\n");
			exit(EX_UNAVAILABLE);
		}
#ifdef DEBUG
		debug_fp = fopen("log", "w");
		c = smfi_main();
		fclose(debug_fp);
		return c;
#else
		if ( fork() == 0 ) {
			return smfi_main();
		}
		return 0;
#endif
	} else {
#ifdef DEBUG
		debug_fp = fopen("log", "w");
#endif
		/* do stream filter action */
		argv += optind;
		if ( *argv ) {
			while ( *argv ) {
				fname = *argv;
				read_lines = 0;
#ifdef MSDOS
				fp = fopen(*argv, "rb");
#else
				fp = fopen(*argv, "r");
#endif
				if ( fp == NULL ) {
					perror(*argv);
					++argv;
					continue;
				}
				file_filter(fp);
				fclose(fp);
				++argv;
			}
		} else {
			file_filter(stdin);
		}
	}
#ifdef DEBUG
	fclose(debug_fp);
#endif
	return 0;
}
